package ru.dbotthepony.kommons.random

import java.util.random.RandomGenerator

/**
 * [Shishua](https://github.com/espadrine/shishua) implementation, consists of this random generator and [ShishuaEngine], which generates the bits
 *
 * Because JVM isn't smart enough, [ShishuaEngine] doesn't get vectorized, and hence it is slow, almost 5 times
 * slower than [GJRAND64Random]
 */
open class ShishuaRandom() : RandomGenerator {
	protected val engine = ShishuaEngine()
	private val values = LongArray(ShishuaEngine.OUTPUT_SIZE)
	private var index = 16
	private val acceptor = ShishuaOutput { System.arraycopy(it, 0, values, 0, it.size) }

	constructor(seed: LongArray) : this() {
		engine.initialize(seed)
	}

	constructor(seed0: Long, seed1: Long, seed2: Long, seed3: Long) : this() {
		engine.initialize(longArrayOf(seed0, seed1, seed2, seed3))
	}

	constructor(seed: Long) : this() {
		val rng = PCG32Random(seed)
		engine.initialize(longArrayOf(rng.nextLong(), rng.nextLong(), rng.nextLong(), rng.nextLong()))
	}

	protected fun round(): Long {
		if (index == 16) {
			index = 0
			engine.step(acceptor, 1)
		}

		return values[index++]
	}

	override fun nextLong(): Long {
		return round()
	}
}
